'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _regenerator = require('babel-runtime/regenerator');

var _regenerator2 = _interopRequireDefault(_regenerator);

var _reactRedux = require('react-redux');

var _effects = require('redux-saga/effects');

var _utils = require('./utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

exports.default = Bundle;


function undeclaredDependencyError(target, property) {
  throw new Error('use of undeclared dependency ' + property);
}

function Bundle(linker, builder) {
  this.builder = builder;
  this.locals = (0, _utils.makeSafeProxy)({}, undeclaredDependencyError);
  this._ = {
    linker: linker,
    bundles: [],
    earlyReducers: [],
    actionReducers: [],
    lateReducers: [],
    views: [],
    defers: [],
    sagas: [],
    sealed: false
  };
}

Bundle.prototype.include = function (builder) {
  this._assertNotSealed();
  var bundle = new Bundle(this._.linker, builder);
  this._.bundles.push(bundle);
  builder(bundle, bundle.locals);
  return bundle.locals;
};

Bundle.prototype.use = function () {
  this._assertNotSealed();

  for (var _len = arguments.length, names = Array(_len), _key = 0; _key < _len; _key++) {
    names[_key] = arguments[_key];
  }

  this._.linker.declareUse(this.locals, names);
};

Bundle.prototype.pack = function () {
  this._assertNotSealed();
  var target = (0, _utils.makeSafeProxy)({}, undeclaredDependencyError);

  for (var _len2 = arguments.length, names = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
    names[_key2] = arguments[_key2];
  }

  this._.linker.declareUse(target, names);
  return target;
};

Bundle.prototype.defineValue = function (name, value) {
  this._assertNotSealed();
  this._.linker.publish('value', name, value);
  this.use(name);
};

Bundle.prototype.defineSelector = function (name, value) {
  this._assertNotSealed();
  this._.linker.publish('selector', name, value);
  this.use(name);
};

Bundle.prototype.defineAction = function (name, actionType) {
  this._assertNotSealed();
  this._.linker.declareActionType(actionType, name);
  this._.linker.publish('action', name, actionType);
  this.use(name);
};

Bundle.prototype.defineView = function (name, selector, view) {
  this._assertNotSealed();
  if (view === undefined) {
    view = selector;
    selector = undefined;
  }
  this.use(name);
  this._.views.push({ name: name, view: view, selector: selector });
};

Bundle.prototype.addReducer = function (name, reducer) {
  this._assertNotSealed();
  if (reducer === undefined) {
    this._.lateReducers.push(name); // name is the reducer function
  } else {
    this.use(name);
    this._.actionReducers.push({ name: name, reducer: reducer });
  }
};

Bundle.prototype.addEarlyReducer = function (reducer) {
  this._assertNotSealed();
  this._.earlyReducers.push(reducer);
};

Bundle.prototype.addLateReducer = function (reducer) {
  this._assertNotSealed();
  this._.lateReducers.push(reducer);
};

Bundle.prototype.addSaga = function (saga) {
  this._assertNotSealed();
  this._.sagas.push(saga);
};

Bundle.prototype.addEnhancer = function (enhancer) {
  this._assertNotSealed();
  this._.linker.addEnhancer(enhancer);
};

Bundle.prototype.defer = function (callback) {
  this._assertNotSealed();
  this._.defers.push(callback);
};

Bundle.prototype.lookup = function (name) {
  return this._.linker.lookup(name);
};

/* TODO: hide this Bundle methods */

Bundle.prototype._assertNotSealed = function () {
  if (this._.sealed) {
    throw new Error('Dynamically calling epic-linker directives is not supported.');
  }
};

Bundle.prototype._linkViews = function () {
  var i;
  // Define and connect views.
  for (i = 0; i < this._.views.length; i += 1) {
    var _$views$i = this._.views[i],
        name = _$views$i.name,
        selector = _$views$i.selector,
        view = _$views$i.view;

    if (selector !== undefined) {
      if (typeof selector === 'string') {
        selector = this.locals[selector];
      }
      if (typeof selector !== 'function') {
        throw new Error('invalid selector for view', name);
      }
      view = (0, _reactRedux.connect)(selector)(view);
    }
    view.displayName = 'View(' + name + ')';
    this._.linker.publish('view', name, view);
  }
  // Define and connect views in included bundles.
  for (i = 0; i < this._.bundles.length; i += 1) {
    this._.bundles[i]._linkViews();
  }
};

Bundle.prototype._runDefers = function () {
  var i;
  // The bundle's defers run first,

  for (var _len3 = arguments.length, args = Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
    args[_key3] = arguments[_key3];
  }

  for (i = 0; i < this._.defers.length; i += 1) {
    var _$defers$i;

    (_$defers$i = this._.defers[i]).call.apply(_$defers$i, [null].concat(args));
  }
  // followed by the defers in included bundles.
  for (i = 0; i < this._.bundles.length; i += 1) {
    var _$bundles$i;

    (_$bundles$i = this._.bundles[i])._runDefers.apply(_$bundles$i, args);
  }
};

Bundle.prototype._earlyReducer = function () {
  var reducers = this._.earlyReducers.concat(this._.bundles.map(function (bundle) {
    return bundle._earlyReducer();
  }));
  // [x1,…,xn].reduce(f, a) = f(f(f(a,x1), …), xn)
  // Use directCompose so that early-reducers added first apply first.
  return reducers.reduce(_utils.reverseCompose, null);
};

Bundle.prototype._buildActionMap = function (actionMap) {
  var i;
  for (i = 0; i < this._.actionReducers.length; i += 1) {
    var _$actionReducers$i = this._.actionReducers[i],
        name = _$actionReducers$i.name,
        reducer = _$actionReducers$i.reducer;

    var actionType = this.locals[name];
    var prevReducer = actionMap.get(actionType);
    actionMap.set(actionType, (0, _utils.reverseCompose)(prevReducer, reducer));
  }
  for (i = 0; i < this._.bundles.length; i += 1) {
    this._.bundles[i]._buildActionMap(actionMap);
  }
};

Bundle.prototype._lateReducer = function () {
  var reducers = this._.lateReducers.concat(this._.bundles.map(function (bundle) {
    return bundle._lateReducer();
  }));
  // [x1,…,xn].reduce(f, a) = f(f(f(a,x1), …), xn)
  // Use directCompose so that late-reducers added first apply last.
  return reducers.reduce(_utils.directCompose, null);
};

Bundle.prototype._saga = function () {
  var i;
  var effects = [];
  for (i = 0; i < this._.sagas.length; i += 1) {
    effects.push((0, _effects.call)(this._.sagas[i]));
  }
  for (i = 0; i < this._.bundles.length; i += 1) {
    effects.push((0, _effects.call)(this._.bundles[i]._saga()));
  }
  return _regenerator2.default.mark(function _callee() {
    return _regenerator2.default.wrap(function _callee$(_context) {
      while (1) {
        switch (_context.prev = _context.next) {
          case 0:
            _context.next = 2;
            return (0, _effects.all)(effects);

          case 2:
          case 'end':
            return _context.stop();
        }
      }
    }, _callee, this);
  });
};

Bundle.prototype._seal = function () {
  var i;
  this._.sealed = true;
  for (i = 0; i < this._.bundles.length; i += 1) {
    this._.bundles[i]._seal();
  }
};