'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _getIterator2 = require('babel-runtime/core-js/get-iterator');

var _getIterator3 = _interopRequireDefault(_getIterator2);

var _map = require('babel-runtime/core-js/map');

var _map2 = _interopRequireDefault(_map);

exports.default = link;

var _redux = require('redux');

var _reduxSaga = require('redux-saga');

var _reduxSaga2 = _interopRequireDefault(_reduxSaga);

var _utils = require('./utils');

var _bundle = require('./bundle');

var _bundle2 = _interopRequireDefault(_bundle);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/*

This file contains a linker for bundles of (redux) actions and reducers,
(redux-saga) selectors and sagas, and (React) views.

A bundle is implemented as a generator that yields linker steps to:

  - include another bundle;
  - declare the use of dependencies defined by other bundles;
  - define an action, selector, or view;
  - add an action reducer or a saga

A bundle's generator function takes as its single argument an object which
the linker populates with the definitions made or used by the bundle.
A shared flat namespace is used for actions, selectors, and views.

Actions are redux actions and have string values.

Selectors are cheap functions that take the global state and extract a limited
view of it.  A selector may be defined in terms of other selectors, and may
also be used in action reducers and sagas.

Multiple action reducers for the same action are currently not supported (an
exception is thrown at link time), but they could be composed in link order.

If a selector match the name of a view name, the view is connected to the
store using the selector.

*/

function link(rootBuilder) {

  // The global namespace map (name → value)
  var globalScope = {};

  // Type map (value|selector|action|view) used to stage injections.
  var typeMap = new _map2.default();

  // Map(action type → action name)
  var nameForActionType = new _map2.default();

  // Enhancers have a flat structure.
  var enhancers = [];

  // 'use' directives are queued and dependency objects are populated after
  // all definitions have taken effect.
  var useQueue = [];

  var linkErrors = [];

  function undefinedNameError(target, property) {
    throw new Error('use of undefined name ' + property);
  }

  function declareActionType(actionType, name) {
    if (nameForActionType.has(actionType)) {
      throw new Error('action type conflict: ' + actionType);
    }
    nameForActionType.set(actionType, name);
  }

  function declareUse(target, names) {
    useQueue.push([target, names]);
  }

  function addEnhancer(enhancer) {
    enhancers.push(enhancer);
  }

  /* Publish a value in the global scope. */
  function publish(type, name, value) {
    if (name in globalScope) {
      throw new Error('linker conflict on ' + name);
    }
    typeMap.set(name, type);
    globalScope[name] = value;
  }

  /* Look up a value in the global scope. */
  function lookup(name) {
    return globalScope[name];
  }

  /* Inject a value into a local scope. */
  function inject(typeFilter, locals, name) {
    if (typeFilter) {
      var type = typeMap.get(name);
      if (-1 === typeFilter.indexOf(type)) {
        return;
      }
    }
    if (name in globalScope) {
      locals[name] = globalScope[name];
    } else {
      throw new Error('undefined dependency: ' + name);
    }
  }

  /* Inject all values in all local scopes. */
  function injectAll(typeFilter) {
    for (var i = 0; i < useQueue.length; i += 1) {
      var dir = useQueue[i],
          locals = dir[0],
          arg = dir[1];
      if (typeof arg === 'string') {
        inject(typeFilter, locals, arg);
      } else if (Array.isArray(arg)) {
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
          for (var _iterator = (0, _getIterator3.default)(arg), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
            var name = _step.value;

            inject(typeFilter, locals, name);
          }
        } catch (err) {
          _didIteratorError = true;
          _iteratorError = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }
          } finally {
            if (_didIteratorError) {
              throw _iteratorError;
            }
          }
        }
      } else {
        throw new Error('invalid use');
      }
    }
  }

  // Call the root builder with a root bundle.
  // This will directly defines all actions and selectors.
  var linker = { publish: publish, declareUse: declareUse, declareActionType: declareActionType, addEnhancer: addEnhancer, lookup: lookup };
  var rootBundle = new _bundle2.default(linker, rootBuilder);
  rootBuilder(rootBundle, rootBundle.locals);

  // Seal the bundles to ensure all linking is done statically.
  rootBundle._seal();

  /* Views can depend on selector definitions, so inject them in a second phase. */
  injectAll(['action', 'selector', 'value']);
  rootBundle._linkViews();
  injectAll('views');

  // Compose the reducer now that all actions have been defined.
  var actionMap = new _map2.default();
  rootBundle._buildActionMap(actionMap);
  var actionReducer = function actionReducer(state, action) {
    if (actionMap.has(action.type)) {
      state = actionMap.get(action.type)(state, action);
    }
    return state;
  };
  var reducer = [rootBundle._earlyReducer(), actionReducer, rootBundle._lateReducer()].reduce(_utils.reverseCompose, null);

  // Compose the enhancers.
  var sagaMiddleware = (0, _reduxSaga2.default)();
  var enhancer = (0, _redux.applyMiddleware)(sagaMiddleware);
  var _iteratorNormalCompletion2 = true;
  var _didIteratorError2 = false;
  var _iteratorError2 = undefined;

  try {
    for (var _iterator2 = (0, _getIterator3.default)(enhancers), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
      var other = _step2.value;

      enhancer = (0, _redux.compose)(enhancer, other);
    }

    // Create the store.
  } catch (err) {
    _didIteratorError2 = true;
    _iteratorError2 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion2 && _iterator2.return) {
        _iterator2.return();
      }
    } finally {
      if (_didIteratorError2) {
        throw _iteratorError2;
      }
    }
  }

  var store = (0, _redux.createStore)(reducer, null, enhancer);

  function finalize() {
    /* Call the deferred callbacks. */
    rootBundle._runDefers.apply(rootBundle, arguments);
  }

  /* Collect the sagas.  The root task is returned, suggested use is:
       start().done.catch(function (error) {
        // notify user that the application has crashed and offer
        // to restart it by calling start() again.
      });
    */
  var rootSaga = rootBundle._saga();
  function start() {
    return sagaMiddleware.run(rootSaga);
  }

  return {
    scope: (0, _utils.makeSafeProxy)(globalScope, undefinedNameError),
    store: store,
    reducer: reducer,
    finalize: finalize,
    start: start
  };
};